from __future__ import annotations

from collections import deque
from contextvars import ContextVar
from typing import (
    Any,
    ContextManager,
    Deque,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Union,
    cast,
)

from di._utils.dag import topsort
from di._utils.execution_planning import SolvedDependantCache, plan_execution
from di._utils.inspect import is_async_gen_callable, is_coroutine_callable
from di._utils.nullcontext import nullcontext
from di._utils.state import ContainerState, LocalScopeContext
from di._utils.task import AsyncTask, SyncTask, Task
from di.exceptions import DuplicateScopeError, SolvingError
from di.executors import DefaultExecutor
from di.types import FusedContextManager
from di.types.dependencies import DependantBase, DependencyParameter
from di.types.executor import AsyncExecutor, SyncExecutor
from di.types.providers import (
    DependencyProvider,
    DependencyProviderType,
    DependencyType,
)
from di.types.scopes import Scope
from di.types.solved import SolvedDependant

Dependency = Any


class Container:
    _context: ContextVar[ContainerState]
    _executor: Union[AsyncExecutor, SyncExecutor]

    def __init__(
        self,
        *,
        execution_scope: Scope = None,
        executor: Optional[Union[AsyncExecutor, SyncExecutor]] = None,
    ) -> None:
        self._context = ContextVar("context")
        state = ContainerState()
        state.cached_values.add_scope("container")
        state.cached_values.set(Container, self, scope="container")
        self._context.set(state)
        self._executor = executor or DefaultExecutor()
        self._execution_scope = execution_scope

    @property
    def _state(self) -> ContainerState:
        return self._context.get()

    @property
    def scopes(self) -> List[Scope]:
        return self._state.scopes

    def enter_global_scope(self, scope: Scope) -> FusedContextManager[None]:
        """Enter a global scope that is share amongst threads and coroutines.

        If you enter a global scope in one thread / coroutine, it will propagate to others.
        """
        return self._state.enter_scope(scope)

    def enter_local_scope(self, scope: Scope) -> FusedContextManager[None]:
        """Enter a local scope that is localized to the current thread or coroutine.

        If you enter a local scope in one thread / coroutine, it will NOT propagate to others.
        """
        if scope in self._state.stacks:
            raise DuplicateScopeError(f"Scope {scope} has already been entered!")
        return LocalScopeContext(self._context, scope)

    def bind(
        self,
        provider: DependantBase[DependencyType],
        dependency: DependencyProviderType[DependencyType],
    ) -> ContextManager[None]:
        """Bind a new dependency provider for a given dependency.

        This can be used as a function (for a permanent bind, cleared when `scope` is exited)
        or as a context manager (the bind will be cleared when the context manager exits).

        Binds are only identified by the identity of the callable and do not take into account
        the scope or any other data from the dependency they are replacing.
        """
        return self._state.bind(provider=provider, dependency=dependency)

    @property
    def binds(self) -> Mapping[DependencyProvider, DependantBase[Any]]:
        return self._state.binds

    def solve(
        self,
        dependency: DependantBase[DependencyType],
    ) -> SolvedDependant[DependencyType]:
        """Solve a dependency.

        Returns a SolvedDependant that can be executed to get the dependency's value.
        """

        # If the SolvedDependant itself is a bind, replace it's dependant
        if dependency.call in self._state.binds:
            dependency = self._state.binds[dependency.call]

        # Mapping of already seen dependants to itself for hash based lookups
        dep_registry: Dict[DependantBase[Any], DependantBase[Any]] = {}

        # DAG mapping dependants to their dependendencies
        dep_dag: Dict[DependantBase[Any], List[DependantBase[Any]]] = {}

        # The same DAG as above but including parameters (inspect.Parameter instances)
        param_graph: Dict[
            DependantBase[Any],
            List[DependencyParameter[DependantBase[Any]]],
        ] = {}

        def get_params(
            dep: DependantBase[Any],
        ) -> List[DependencyParameter[DependantBase[Any]]]:
            # get parameters and swap them out w/ binds when they
            # exist as a bound value
            params = dep.get_dependencies().copy()
            for idx, param in enumerate(params):
                assert param.dependency.call is not None
                if param.dependency.call in self._state.binds:
                    params[idx] = DependencyParameter[Any](
                        dependency=self._state.binds[param.dependency.call],
                        parameter=param.parameter,
                    )
            return params

        def check_equivalent(dep: DependantBase[Any]) -> None:
            if dep in dep_registry and dep.scope != dep_registry[dep].scope:
                raise SolvingError(
                    f"The dependencies {dep} and {dep_registry[dep]}"
                    " have the same lookup (__hash__ and __eq__) but have different scopes"
                    f" ({dep.scope} and {dep_registry[dep].scope} respectively)"
                    " This is likely a mistake, but you can override this behavior by either:"
                    "\n  1. Wrapping the function in another function or subclassing the class such"
                    " that they now are considered different dependencies"
                    "\n  2. Use a custom implemetnation of DependantBase that changes the meaning"
                    " or computation of __hash__ and/or __eq__"
                )

        # Do a DFS of the DAG checking constraints along the way
        q: Deque[DependantBase[Any]] = deque([dependency])
        while q:
            dep = q.popleft()
            if dep in dep_registry:
                check_equivalent(dep)
            else:
                dep_registry[dep] = dep
                params = get_params(dep)
                param_graph[dep] = params
                dep_dag[dep] = []
                for param in params:
                    subdep = param.dependency
                    dep_dag[dep].append(subdep)
                    if subdep not in dep_registry:
                        q.append(subdep)
        # The whole concept of Tasks is something that can perhaps be cleaned up / optimized
        # The main reason to have it is to save some computation at runtime
        # but currently that is just checking if the dependendency is sync or async
        # So perhaps that can be done in a simpler manner
        tasks = self._build_tasks(param_graph, topsort(dep_dag))
        dependency_dag = {dep: set(subdeps) for dep, subdeps in dep_dag.items()}
        container_cache = SolvedDependantCache(
            dependency_dag=dependency_dag, tasks=tasks
        )
        return SolvedDependant(
            dependency=dependency,
            dag=param_graph,
            container_cache=container_cache,
        )

    def _build_tasks(
        self,
        dag: Dict[
            DependantBase[Any],
            List[DependencyParameter[DependantBase[Any]]],
        ],
        topsorted: List[DependantBase[Any]],
    ) -> Dict[DependantBase[Any], Union[AsyncTask[Any], SyncTask[Any]]]:
        tasks: Dict[DependantBase[Any], Union[AsyncTask[Any], SyncTask[Any]]] = {}
        for dep in reversed(topsorted):
            task_dependencies: List[DependencyParameter[Task[Any]]] = [
                DependencyParameter(
                    dependency=tasks[param.dependency], parameter=param.parameter
                )
                for param in dag[dep]
            ]

            if is_async_gen_callable(dep.call) or is_coroutine_callable(dep.call):
                tasks[dep] = AsyncTask(dependant=dep, dependencies=task_dependencies)
            else:
                tasks[dep] = SyncTask(dependant=dep, dependencies=task_dependencies)
        return tasks

    def _update_cache(
        self,
        results: Dict[DependantBase[Any], Any],
        to_cache: Iterable[DependantBase[Any]],
    ) -> None:
        for dep in to_cache:
            self._state.cached_values.set(
                dep.call,  # type: ignore[arg-type]
                results[dep],
                scope=dep.scope,
            )

    def execute_sync(
        self,
        solved: SolvedDependant[DependencyType],
        *,
        validate_scopes: bool = True,
        values: Optional[Mapping[DependencyProvider, Any]] = None,
    ) -> DependencyType:
        """Execute an already solved dependency.

        This method is synchronous and uses a synchronous executor,
        but the executor may still be able to execute async dependencies.

        If you are not dynamically changing scopes, you can run once with `validate_scopes=True`
        and then disable scope validation in subsequent runs with `validate_scope=False`.
        """
        cm: FusedContextManager[None]
        if self._execution_scope in self.scopes:
            cm = nullcontext()
        else:
            cm = self.enter_local_scope(self._execution_scope)
        with cm:
            results, leaf_tasks, to_cache = plan_execution(
                self._state, solved, validate_scopes=validate_scopes, values=values
            )
            if not hasattr(self._executor, "execute_sync"):  # pragma: no cover
                raise TypeError(
                    "execute_sync requires an executor implementing the SyncExecutor protocol"
                )
            executor = cast(SyncExecutor, self._executor)
            if leaf_tasks:
                executor.execute_sync(leaf_tasks)
            self._update_cache(results, to_cache)
            return results[solved.dependency]  # type: ignore[no-any-return]

    async def execute_async(
        self,
        solved: SolvedDependant[DependencyType],
        *,
        validate_scopes: bool = True,
        values: Optional[Mapping[DependencyProvider, Any]] = None,
    ) -> DependencyType:
        """Execute an already solved dependency.

        If you are not dynamically changing scopes, you can run once with `validate_scopes=True`
        and then disable scope validation in subsequent runs with `validate_scope=False`.
        """
        cm: FusedContextManager[None]
        if self._execution_scope in self.scopes:
            cm = nullcontext()
        else:
            cm = self.enter_local_scope(self._execution_scope)
        async with cm:
            results, leaf_tasks, to_cache = plan_execution(
                self._state, solved, validate_scopes=validate_scopes, values=values
            )
            if not hasattr(self._executor, "execute_async"):  # pragma: no cover
                raise TypeError(
                    "execute_async requires an executor implementing the AsyncExecutor protocol"
                )
            executor = cast(AsyncExecutor, self._executor)
            if leaf_tasks:
                await executor.execute_async(leaf_tasks)
            self._update_cache(results, to_cache)
            return results[solved.dependency]  # type: ignore[no-any-return]



from __future__ import annotations

import functools
import typing
from collections import deque
from contextlib import AsyncExitStack, asynccontextmanager, contextmanager
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Deque,
    Dict,
    Generic,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Tuple,
    Union,
    cast,
)

from di._utils.inspect import is_async_gen_callable, is_gen_callable
from di._utils.state import ContainerState
from di.exceptions import IncompatibleDependencyError
from di.types.dependencies import DependantBase, DependencyParameter
from di.types.executor import AsyncTaskInfo, SyncTaskInfo, TaskInfo
from di.types.providers import (
    AsyncGeneratorProvider,
    CallableProvider,
    CoroutineProvider,
    DependencyType,
    GeneratorProvider,
)


class ExecutionState(typing.NamedTuple):
    container_state: ContainerState
    results: Dict[DependantBase[Any], Any]
    dependency_counts: MutableMapping[DependantBase[Any], int]
    dependants: Mapping[DependantBase[Any], Iterable[Task[Any]]]


class Task(Generic[DependencyType]):
    __slots__ = ("dependant", "dependencies")

    def __init__(
        self,
        dependant: DependantBase[DependencyType],
        dependencies: List[DependencyParameter[Task[Any]]],
    ) -> None:
        self.dependant = dependant
        self.dependencies = dependencies

    def compute(
        self,
        state: ExecutionState,
    ) -> Union[Awaitable[Iterable[Optional[TaskInfo]]], Iterable[Optional[TaskInfo]]]:
        """Compute this dependency within the context of the current state"""
        raise NotImplementedError

    def as_executor_task(self, state: ExecutionState) -> TaskInfo:
        """Bind the state and return either an AsyncTaskInfo or SyncTaskInfo"""
        raise NotImplementedError

    def gather_params(
        self, results: Dict[DependantBase[Any], Any]
    ) -> Tuple[List[Any], Dict[str, Any]]:
        """Gather all parameters (aka *args and **kwargs) needed for computation"""
        positional: List[Any] = []
        keyword: Dict[str, Any] = {}
        for dep in self.dependencies:
            if dep.parameter is not None:
                if dep.parameter.kind is dep.parameter.kind.POSITIONAL_ONLY:
                    positional.append(results[dep.dependency.dependant])
                else:
                    keyword[dep.parameter.name] = results[dep.dependency.dependant]
        return positional, keyword

    def gather_new_tasks(self, state: ExecutionState) -> Iterable[Optional[TaskInfo]]:
        """Look amongst our dependants to see if any of them are now dependency free"""
        new_tasks: Deque[Optional[TaskInfo]] = deque()
        for dependant in state.dependants[self.dependant]:
            count = state.dependency_counts[dependant.dependant]
            if count == 1:
                # this dependant has no further dependencies, so we can compute it now
                new_tasks.append(dependant.as_executor_task(state))
                # pop it from dependency counts so that we can
                # tell when we've satisfied all dependencies below
                state.dependency_counts.pop(dependant.dependant)
            else:
                state.dependency_counts[dependant.dependant] = count - 1
        # also pop ourselves if we have no deps
        if state.dependency_counts.get(self.dependant, -1) == 0:
            state.dependency_counts.pop(self.dependant)
        if not state.dependency_counts:
            # all dependencies are taken care of, insert sentinel None value
            new_tasks.append(None)
        return new_tasks

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({repr(self.dependant)}, {repr(self.dependencies)})"


class AsyncTask(Task[DependencyType]):
    __slots__ = ("is_generator", "call")

    def __init__(
        self,
        dependant: DependantBase[DependencyType],
        dependencies: List[DependencyParameter[Task[Any]]],
    ) -> None:
        super().__init__(dependant, dependencies)
        assert self.dependant.call is not None
        self.call = self.dependant.call
        self.is_generator = is_async_gen_callable(self.call)

    def as_executor_task(self, state: ExecutionState) -> AsyncTaskInfo:
        return AsyncTaskInfo(
            dependant=self.dependant,
            task=functools.partial(self.compute, state),
        )

    async def compute(
        self,
        state: ExecutionState,
    ) -> Iterable[Optional[TaskInfo]]:
        args, kwargs = self.gather_params(state.results)

        call = self.call
        if self.is_generator:
            stack = state.container_state.stacks[self.dependant.scope]
            if not isinstance(stack, AsyncExitStack):
                raise IncompatibleDependencyError(
                    f"The dependency {self.dependant} is an awaitable dependency"
                    f" and canot be used in the sync scope {self.dependant.scope}"
                )
            if TYPE_CHECKING:
                call = cast(AsyncGeneratorProvider[DependencyType], call)
            state.results[self.dependant] = await stack.enter_async_context(
                asynccontextmanager(call)(*args, **kwargs)
            )
        else:
            if TYPE_CHECKING:
                call = cast(CoroutineProvider[DependencyType], call)
            state.results[self.dependant] = await call(*args, **kwargs)

        return self.gather_new_tasks(state)


class SyncTask(Task[DependencyType]):
    __slots__ = ("is_generator", "call")

    def __init__(
        self,
        dependant: DependantBase[DependencyType],
        dependencies: List[DependencyParameter[Task[Any]]],
    ) -> None:
        super().__init__(dependant, dependencies)
        assert self.dependant.call is not None
        self.call = self.dependant.call
        self.is_generator = is_gen_callable(self.call)

    def as_executor_task(self, state: ExecutionState) -> SyncTaskInfo:
        return SyncTaskInfo(
            dependant=self.dependant,
            task=functools.partial(self.compute, state),
        )

    def compute(
        self,
        state: ExecutionState,
    ) -> Iterable[Optional[TaskInfo]]:

        args, kwargs = self.gather_params(state.results)

        call = self.call
        if self.is_generator:
            if TYPE_CHECKING:
                call = cast(GeneratorProvider[DependencyType], call)
            stack = state.container_state.stacks[self.dependant.scope]
            state.results[self.dependant] = stack.enter_context(
                contextmanager(call)(*args, **kwargs)
            )
        else:
            if TYPE_CHECKING:
                call = cast(CallableProvider[DependencyType], call)
            state.results[self.dependant] = call(*args, **kwargs)

        return self.gather_new_tasks(state)



from di.types import FusedContextManager


class nullcontext(FusedContextManager[None]):
    """Backport of contextlib.nullcontext"""

    def __enter__(self) -> None:
        return None

    def __exit__(self, *excinfo) -> bool:  # type: ignore
        return False

    async def __aenter__(self) -> None:
        return None

    async def __aexit__(self, *excinfo) -> bool:  # type: ignore
        return False



import sys
from typing import Callable, cast

vesion: Callable[[str], str]
if sys.version_info < (3, 8):
    from importlib_metadata import PackageNotFoundError, version

    version = cast(Callable[[str], str], version)
else:
    from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("di")
except PackageNotFoundError:
    __version__ = "0.0.0"


from di.container import Container  # noqa: E402
from di.dependant import (  # noqa: E402
    CallableClassDependant,
    Dependant,
    JoinedDependant,
)
from di.executors import (  # noqa: E402
    ConcurrentAsyncExecutor,
    ConcurrentSyncExecutor,
    DefaultExecutor,
    SimpleAsyncExecutor,
    SimpleSyncExecutor,
)
from di.params import Depends  # noqa: E402

__all__ = (
    "Container",
    "Dependant",
    "JoinedDependant",
    "CallableClassDependant",
    "Depends",
    "DefaultExecutor",
    "ConcurrentAsyncExecutor",
    "ConcurrentSyncExecutor",
    "SimpleAsyncExecutor",
    "SimpleSyncExecutor",
)



from __future__ import annotations

import dataclasses
import typing
from typing import Any, Generic, List, Mapping

from di.types.dependencies import DependantBase, DependencyParameter
from di.types.providers import DependencyType

Dependency = Any


@dataclasses.dataclass
class SolvedDependant(Generic[DependencyType]):
    """Representation of a fully solved dependency as DAG"""

    dependency: DependantBase[DependencyType]
    dag: Mapping[DependantBase[Any], List[DependencyParameter[DependantBase[Any]]]]
    # container_cache can be used by the creating container to store data that is tied
    # to the SolvedDependant
    container_cache: typing.Any = None

    def get_flat_subdependants(self) -> List[DependantBase[Any]]:
        """Get an exhaustive list of all of the dependencies of this dependency,
        in no particular order.
        """
        return list(self.dag.keys() - {self.dependency})



from __future__ import annotations

from typing import Dict, Generic, Hashable, List, TypeVar

from di.exceptions import DuplicateScopeError
from di.types.scopes import Scope

KT = TypeVar("KT", bound=Hashable)
VT = TypeVar("VT")


class ScopeMap(Generic[KT, VT]):
    """Mapping-like structure to hold cached values.

    The important aspect of this is that we may have thousands of dependencies, but generally only a couple (<10) scopes.
    So this is designed to operate in ~ O(S) time, where S is the number of scopes.
    Imporantly, when we enter or exit a scope (which can happen very frequently, e.g. requests in a web framework),
    no iteration is required: we can simply pop all keys in that scope.

    The interface is *almost* a Mapping[KT, VT] but not quite: some operations like set / __setitem__ require a key, a value *and* a scope.
    We could implement ScopeMap[key, scope] = value, but that can lead to a bit of confusion since we want lookups like ScopMap[key].
    Additionally, operations are not O(1) but rather O(S).
    So we choose to use method names like get() and set() to deal with the API differences and indicate that operations are more expensive than O(1).
    This is also similar to collections.ChainMap, except that we want "named" nodes (Scopes) and ChainMap uses a plain list.
    ChainMap also doesn't allow you to set values anywhere but the left mapping, and we need to set values in arbitrary mappings.
    """

    __slots__ = ("mappings",)

    mappings: Dict[Scope, Dict[KT, VT]]

    def __init__(self) -> None:
        self.mappings = {}

    def set(self, key: KT, value: VT, *, scope: Scope) -> None:
        self.mappings[scope][key] = value

    def add_scope(self, scope: Scope) -> None:
        if scope in self.mappings:
            raise DuplicateScopeError(f"The scope {scope} already exists!")
        self.mappings[scope] = {}

    def pop_scope(self, scope: Scope) -> None:
        self.mappings.pop(scope)

    def copy(self) -> ScopeMap[KT, VT]:
        new = type(self)()
        new.mappings = self.mappings.copy()
        return new

    def __repr__(self) -> str:  # pragma: no cover, used only for debugging
        values: List[str] = []
        for scope, mapping in self.mappings.items():
            for k, v in mapping.items():
                values.append(f'{repr(k)}: {repr(v)} @ scope="{scope}"')
        return "{" + ", ".join(values) + "}"



class DependencyInjectionException(Exception):
    """Base exception for this library"""

    pass


class WiringError(DependencyInjectionException):
    """Raised when wiring (introspection into types) failed"""

    pass


class UnknownScopeError(DependencyInjectionException):
    """Raised when a dependency to be executed has an unknown Scope"""

    pass


class DuplicateScopeError(DependencyInjectionException):
    """Raised when enter_local_scope() or enter_global_scope() is called with an existing scope"""

    pass


class CircularDependencyError(DependencyInjectionException):
    """Raised when a dependency cycle is detected"""

    pass


class ScopeViolationError(DependencyInjectionException):
    """Raised when Scope layering is violated.
    Using pytests' Scopes as an example, if A has "session" Scope and B has "function" Scope,
    A cannot depend on B (in fact, pytest will also throw an error).
    """

    pass


class SolvingError(DependencyInjectionException):
    """Raised when there is an issue solving, for example if a dependency appears twice with different scopes"""

    pass


class IncompatibleDependencyError(DependencyInjectionException):
    """Raised when an async context manager dependency is executed in a sync Scope"""

    pass



from collections import deque
from typing import (
    AbstractSet,
    Any,
    Deque,
    Dict,
    Iterable,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Set,
    Tuple,
)

from di._utils import scope_validation
from di._utils.state import ContainerState
from di._utils.task import ExecutionState, Task
from di.types.dependencies import DependantBase
from di.types.executor import TaskInfo
from di.types.providers import DependencyProvider
from di.types.solved import SolvedDependant

Dependency = Any


class ExecutionPlan(NamedTuple):
    """Cache of the execution plan, stored within SolvedDependantCache"""

    # the cache_key is used to compare execution plans and know if we need to re-create it
    # for a given SolvedDependant if the cache_key is equal, the plans will be equal
    cache_key: AbstractSet[DependantBase[Any]]
    # mapping of which dependencies require computation for each dependant
    # it is possible that some of them were passed by value or cached
    # and hence do not require computation
    dependant_dag: Mapping[DependantBase[Any], Iterable[Task[Any]]]
    # count of how many uncompomputed dependencies each dependant has
    # this is used to keep track of what dependencies are now available for computation
    dependency_counts: Dict[DependantBase[Any], int]
    leaf_tasks: Iterable[Task[Any]]  # these are the tasks w/ no dependencies


class SolvedDependantCache(NamedTuple):
    """Private data that the Container attaches to SolvedDependant"""

    tasks: Mapping[DependantBase[Any], Task[Any]]
    dependency_dag: Mapping[DependantBase[Any], Set[DependantBase[Any]]]
    execution_plan: Optional[ExecutionPlan] = None


def plan_execution(
    state: ContainerState,
    solved: SolvedDependant[Any],
    *,
    validate_scopes: bool = True,
    values: Optional[Mapping[DependencyProvider, Any]] = None,
) -> Tuple[
    Dict[DependantBase[Any], Any],
    List[TaskInfo],
    Iterable[DependantBase[Any]],
]:
    """Re-use or create an ExecutionPlan"""
    # This function is a hot loop
    # It is run for every execution, and even with the cache it can be a bottleneck
    # This would be the best place to Cythonize or otherwise take more drastic
    # measures to improve performance

    user_values = values or {}
    if validate_scopes:
        scope_validation.validate_scopes(state.scopes, solved)

    if type(solved.container_cache) is not SolvedDependantCache:
        raise TypeError(
            "This SolvedDependant was not created by this Container"
        )  # pragma: no cover

    solved_dependency_cache: SolvedDependantCache = solved.container_cache

    # dump all cached values into one mapping for faster lookups
    # this relies on the fact that:
    # (1) dict.update is fast
    # (2) we will be doing a lot of lookups (one for each key in the DAG)
    # (3) this converts an iteration over scopes into a dict lookup
    # The cost of this is extra memory: we are building up `cache` w/ values that will never be used
    # And then we are throwing `cache` away when we exit this function
    cache: Dict[DependencyProvider, Any] = {}
    for mapping in state.cached_values.mappings.values():
        cache.update(mapping)
    cache.update(user_values)

    to_cache: Deque[DependantBase[Any]] = deque()
    execution_scope = state.scopes[-1]

    results: Dict[DependantBase[Any], Any] = {}

    # For each dependency, check if we can use a cached value or a user supplied value
    # If so, put that value into our results dict
    for dep in solved.dag:
        call = dep.call
        if call in cache:
            if call in user_values:
                results[dep] = user_values[call]
            elif dep.share:
                results[dep] = cache[call]
        else:
            # skip caching the innermost scope since it would be dropped
            # right after saving the value
            if dep.share and dep.scope != execution_scope:
                # mark this dependency to have it's value stored in the cache
                # after we are done w/ this execution
                to_cache.append(dep)

    execution_plan_cache_key = results.keys()

    execution_plan = solved_dependency_cache.execution_plan

    # Check if we can re-use the existing plan (if any) or create a new one
    if execution_plan is None or execution_plan.cache_key != execution_plan_cache_key:
        # Build a DAG of Tasks that we actually need to execute
        # this allows us to prune subtrees that will come
        # from pre-computed values (values paramter or cache)
        unvisited = deque([solved.dependency])
        dependency_counts: Dict[DependantBase[Any], int] = {}
        dependant_dag: Dict[DependantBase[Any], Deque[Task[Any]]] = {
            dep: deque() for dep in solved_dependency_cache.dependency_dag
        }
        while unvisited:
            dep = unvisited.pop()
            if dep in dependency_counts:
                continue
            # if the dependency is cached or was provided by value
            # we don't need to compute it or any of it's dependencies
            if dep not in results:
                # otherwise, we add it to our DAG and visit it's children
                dependency_counts[dep] = 0
                for subdep in solved_dependency_cache.dependency_dag[dep]:
                    if subdep not in results:
                        dependant_dag[subdep].append(solved_dependency_cache.tasks[dep])
                        dependency_counts[dep] += 1
                        if subdep not in dependency_counts:
                            unvisited.append(subdep)

        execution_plan = ExecutionPlan(
            # use frozenset to hold a shallow copy of the dict keys
            cache_key=frozenset(execution_plan_cache_key),
            dependant_dag=dependant_dag,
            dependency_counts=dependency_counts,
            leaf_tasks=[
                solved_dependency_cache.tasks[dep]
                for dep, count in dependency_counts.items()
                if count == 0
            ],
        )

        solved.container_cache = SolvedDependantCache(
            tasks=solved_dependency_cache.tasks,
            dependency_dag=solved_dependency_cache.dependency_dag,
            execution_plan=execution_plan,
        )

    execution_state = ExecutionState(
        container_state=state,
        results=results,
        dependency_counts=execution_plan.dependency_counts.copy(),
        dependants=execution_plan.dependant_dag,
    )

    return (
        results,
        [t.as_executor_task(execution_state) for t in execution_plan.leaf_tasks],
        to_cache,
    )
